import {
	encodeBase32LowerCaseNoPadding,
	encodeHexLowerCase,
} from "@oslojs/encoding";
import { sha256 } from "@oslojs/crypto/sha2";
import type { RequestEvent } from "@sveltejs/kit";
import { eq } from "drizzle-orm";
import { Temporal, toTemporalInstant } from "@js-temporal/polyfill";

import { db } from "./db";
import { users, sessions } from "./schema";
import type { User } from "./user";

const sessionExpirySeconds = 30 * 86400;
const sessionExpiry = new Temporal.Duration(0, 0, 0, 0, 0, 0, sessionExpirySeconds);
const sessionRenewal = new Temporal.Duration(0, 0, 0, 0, 0, 0, sessionExpirySeconds/2);

export async function validateSessionToken(
	token: string,
): Promise<SessionValidationResult> {
	const sessionId = encodeHexLowerCase(sha256(new TextEncoder().encode(token)));

	const row = await db.query.sessions.findFirst({
		with: {
			user: true,
		},
		where: (sessions, { eq }) => eq(sessions.id, sessionId),
	});

	if (!row) {
		return { session: null, user: null };
	}
	const session: Session = {
		id: row.id,
		userId: row.userId,
		expiresAt: row.expiresAt,
	};

	const user: User = {
		id: row.userId,
		username: row.user.username,
		displayName: row.user.displayName || undefined,
	};

	const now = Temporal.Now.instant();
	const expiresAt = toTemporalInstant.apply(session.expiresAt);

	if (Temporal.Instant.compare(now, expiresAt) > 0) {
		await db.delete(sessions).where(eq(sessions.id, session.id));
		return { session: null, user: null };
	}

	if (Temporal.Instant.compare(now.add(sessionRenewal), expiresAt) > 0) {
		session.expiresAt = new Date(now.add(sessionExpiry).epochMilliseconds);
		await db.update(sessions)
			.set({ expiresAt: session.expiresAt })
			.where(eq(sessions.id, session.id));
	}

	return { session, user };
}

export async function invalidateSession(sessionId: string): Promise<void> {
	await db.delete(sessions).where(eq(sessions.id, sessionId));
}

export async function invalidateUserSessions(userId: number): Promise<void> {
	await db.delete(sessions).where(eq(sessions.userId, userId));
}

export function setSessionTokenCookie(
	event: RequestEvent,
	token: string,
	expiresAt: Date,
): void {
	event.cookies.set("session", token, {
		httpOnly: true,
		path: "/",
		secure: import.meta.env.PROD,
		sameSite: "lax",
		expires: expiresAt,
	});
}

export function deleteSessionTokenCookie(event: RequestEvent): void {
	event.cookies.set("session", "", {
		httpOnly: true,
		path: "/",
		secure: import.meta.env.PROD,
		sameSite: "lax",
		maxAge: 0,
	});
}

export function generateSessionToken(): string {
	const tokenBytes = new Uint8Array(20);
	crypto.getRandomValues(tokenBytes);
	const token = encodeBase32LowerCaseNoPadding(tokenBytes).toLowerCase();
	return token;
}

export async function createSession(token: string, userId: number): Promise<Session> {
	const sessionId = encodeHexLowerCase(sha256(new TextEncoder().encode(token)));
	const session: Session = {
		id: sessionId,
		userId,
		expiresAt: new Date(
			Temporal.Now.instant().add(sessionExpiry).epochMilliseconds,
		),
	};
	await db.insert(sessions).values(session);
	return session;
}

export interface Session {
	id: string;
	expiresAt: Date;
	userId: number;
}

type SessionValidationResult =
	| { session: Session; user: User }
	| { session: null; user: null };
