import { fail, redirect } from "@sveltejs/kit";


import { RefillingTokenBucket, Throttler } from "$lib/server/rate-limit";
import type { Actions, PageServerLoad, PageServerLoadEvent, RequestEvent } from "./$types";
import { getUserFromUsername, getUserPasswordHash } from "$lib/server/user";
import { verifyPasswordHash } from "$lib/server/password";
import { createSession, generateSessionToken, setSessionTokenCookie } from "$lib/server/session";

const throttler = new Throttler<number>([0, 1, 2, 4, 8, 16, 30, 60, 180, 300]);
const ipBucket = new RefillingTokenBucket<string>(20, 1);

export const load: PageServerLoad = (event: PageServerLoadEvent) => {
	if (event.locals.session !== null && event.locals.user !== null) {
		redirect(303, "/");
	}
	return {};
};

export const actions: Actions = {
	default: action
};

async function action(event: RequestEvent) {
	// TODO: Assumes X-Forwarded-For is always included.
	const clientIP = event.request.headers.get("X-Forwarded-For");
	if (clientIP !== null && !ipBucket.check(clientIP, 1)) {
		return fail(429, {
			message: "Too many requests",
			email: ""
		});
	}

	const formData = await event.request.formData();
	const username = formData.get("username");
	const password = formData.get("password");
	if (typeof username !== "string" || typeof password !== "string") {
		return fail(400, {
			message: "Invalid or missing fields",
			username: ""
		});
	}
	if (username === "" || password === "") {
		return fail(400, {
			message: "Please enter your username and password.",
			username: username
		});
	}
	const user = await getUserFromUsername(username);
	if (user === null) {
		return fail(400, {
			message: "Account does not exist",
			username: username
		});
	}
	if (clientIP !== null && !ipBucket.consume(clientIP, 1)) {
		return fail(429, {
			message: "Too many requests",
			username: ""
		});
	}
	if (!throttler.consume(user.id)) {
		return fail(429, {
			message: "Too many requests",
			username: ""
		});
	}
	const passwordHash = await getUserPasswordHash(user.id);
	if (passwordHash === null) {
		return fail(400, {
			message: "Account locked",
			username
		})
	}
	const validPassword = await verifyPasswordHash(passwordHash, password);
	if (!validPassword) {
		return fail(400, {
			message: "Invalid password",
			username
		});
	}
	throttler.reset(user.id);
	const sessionToken = generateSessionToken();
	const session = await createSession(sessionToken, user.id);
	setSessionTokenCookie(event, sessionToken, session.expiresAt);
	return redirect(303, "/");
}